/*
 *  uri.c
 *
 *  PostgreSQL Base Type URI (RFC 2396)
 *
 *
 */

#include "postgres.h"
#include "fmgr.h"
#include "funcapi.h"
#include "utils/builtins.h"
#include "libpq/pqformat.h"
#include "mb/pg_wchar.h"

#include "liburi/uri.h"

/*
 *  Commonly used code snippets
 */

static
text *uriString_uriText(char *uriString)
{
    if(uriString) {
	int len = strlen(uriString);
	text *uriText = (text *) palloc(VARHDRSZ + len);
	VARATT_SIZEP(uriText) = VARHDRSZ + len;
	memcpy(VARDATA(uriText),uriString,len);
	return uriText;
    }
    return NULL;
}

static
char *uriText_uriString(text *uriText)
{
    if(uriText) {
	char *uriString = (char *) palloc(VARSIZE(uriText) - VARHDRSZ + 1);
	memcpy(uriString,VARDATA(uriText),VARSIZE(uriText) - VARHDRSZ);
	uriString[VARSIZE(uriText) - VARHDRSZ] = '\0';
	return uriString;
    }
    return NULL;
}

/*
 *  Input / Output functions
 */

PG_FUNCTION_INFO_V1(uri_in);
Datum uri_in(PG_FUNCTION_ARGS)
{
    char        *uriString = PG_GETARG_CSTRING(0);
    text        *uriText;
    uri_t       *uriStruct;
    int         len;

    if(uriString && *uriString) uriString = pg_do_encoding_conversion(uriString,strlen(uriString),GetDatabaseEncoding(),PG_LATIN1);
    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(uriString,strlen(uriString));

    if(uriStruct == NULL || uri_consistent(uriStruct) == 0)
        ereport(
            ERROR,
            (errcode(ERRCODE_INVALID_TEXT_REPRESENTATION),
            errmsg(
                "invalid input syntax for uri: \"%s\" - %s",
                uriString,
                uri_errstr
            ))
        );

    uriString = uri_unescape(uri_uri(uriStruct));
    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(uri_out);
Datum uri_out(PG_FUNCTION_ARGS)
{
    text        *uriText = PG_GETARG_TEXT_P(0);
    char        *uriString;
    uri_t       *uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_uri(uriStruct);
    PG_RETURN_CSTRING(pstrdup(uriString));
}

PG_FUNCTION_INFO_V1(uri_recv);
Datum uri_recv(PG_FUNCTION_ARGS)
{
        StringInfo      buf = (StringInfo) PG_GETARG_POINTER(0);
        text       *result;
        char       *str;
        int                     nbytes;

        str = pq_getmsgtext(buf, buf->len - buf->cursor, &nbytes);
        result = (text *) palloc(nbytes + VARHDRSZ);
        VARATT_SIZEP(result) = nbytes + VARHDRSZ;
        memcpy(VARDATA(result), str, nbytes);
        pfree(str);
        PG_RETURN_TEXT_P(result);
}

PG_FUNCTION_INFO_V1(uri_send);
Datum uri_send(PG_FUNCTION_ARGS)
{
        text       *t = PG_GETARG_TEXT_P(0);
        StringInfoData buf;

        pq_begintypsend(&buf);
        pq_sendtext(&buf, VARDATA(t), VARATT_SIZEP(t) - VARHDRSZ);
        PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
}

/*
 *  Cast Functions
 */

PG_FUNCTION_INFO_V1(uri_text);
Datum uri_text(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString = (char *) DirectFunctionCall1(uri_out,(Datum) uriText);

    PG_FREE_IF_COPY(uriText,0);
    uriText = uriString_uriText(uriString);
    pfree(uriString);

    PG_RETURN_TEXT_P(uriText);
}

PG_FUNCTION_INFO_V1(text_uri);
Datum text_uri(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString = uriText_uriString(uriText);
    char	*uriTemp;
    
    PG_FREE_IF_COPY(uriText,0);
    uriText = (text *) DirectFunctionCall1(uri_in,(Datum) uriString);
    pfree(uriString);

    PG_RETURN_TEXT_P(uriText);
}

PG_FUNCTION_INFO_V1(uri_uri_table);
Datum uri_uri_table(PG_FUNCTION_ARGS)
{
    text		*uriText = PG_GETARG_TEXT_P(0);
    uri_t		*uriStruct;
    char		*result[8];

    TupleDesc           tupdesc;
    AttInMetadata       *attinmeta;
    HeapTuple           tuple;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);
    
    if(!uriStruct)
	PG_RETURN_NULL();

    result[0] = uri_scheme(uriStruct);
    result[1] = uri_user(uriStruct);
    result[2] = uri_passwd(uriStruct);
    result[3] = uri_host(uriStruct);
    result[4] = uri_port(uriStruct);
    result[5] = uri_path(uriStruct);
    result[6] = uri_query(uriStruct);
    result[7] = uri_frag(uriStruct);

    tupdesc = RelationNameGetTupleDesc("uri_table");
    attinmeta = TupleDescGetAttInMetadata(tupdesc);
    tuple = BuildTupleFromCStrings(attinmeta, result);

    pfree(tupdesc);
    pfree(attinmeta);

    PG_RETURN_DATUM(HeapTupleGetDatum(tuple));
}

PG_FUNCTION_INFO_V1(uri_table_uri);
Datum uri_table_uri(PG_FUNCTION_ARGS)
{
    HeapTupleHeader	tuple = PG_GETARG_HEAPTUPLEHEADER(0);
    uri_t		*uriStruct = uri_object("http://user:pass@host:80/path;params?query#frag",47);
    text		*uriResult;
    bool		isnull;
    Datum		param;
    
    param = GetAttributeByName(tuple, "scheme", &isnull);
    uri_scheme_set(uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));

    param = GetAttributeByName(tuple, "username", &isnull);
    uri_user_set  (uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));

    param = GetAttributeByName(tuple, "passwd", &isnull);
    uri_passwd_set(uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));

    param = GetAttributeByName(tuple, "host", &isnull);
    uri_host_set  (uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));

    param = GetAttributeByName(tuple, "port", &isnull);
    uri_port_set  (uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));

    param = GetAttributeByName(tuple, "path", &isnull);
    uri_path_set  (uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));

    param = GetAttributeByName(tuple, "query", &isnull);
    uri_query_set (uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));

    param = GetAttributeByName(tuple, "frag", &isnull);
    uri_frag_set  (uriStruct,isnull ? NULL : uriText_uriString(DatumGetTextP(param)));
    
    uriResult = uriString_uriText(uri_uri(uriStruct));
    
    if(uriStruct->scheme)	pfree(uriStruct->scheme);
    if(uriStruct->user)		pfree(uriStruct->user);
    if(uriStruct->passwd)	pfree(uriStruct->passwd);
    if(uriStruct->host)		pfree(uriStruct->host);
    if(uriStruct->port)		pfree(uriStruct->port);
    if(uriStruct->path)		pfree(uriStruct->path);
    if(uriStruct->query)	pfree(uriStruct->query);
    if(uriStruct->frag)		pfree(uriStruct->frag);
    
    if((uriResult == NULL) || (VARSIZE(uriResult) == VARHDRSZ))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriResult);
}

/*
 *  Hash Operator Class Functions
 */

PG_FUNCTION_INFO_V1(uri_hash);
Datum uri_hash(PG_FUNCTION_ARGS)
{
    text                *arg;
    int32               result;

    arg = PG_GETARG_TEXT_P(0);
    result = hash_any(VARDATA(arg), VARSIZE(arg) - VARHDRSZ);
    PG_FREE_IF_COPY(arg,0);
    
    return result;
}

/*
 *  Utility Functions
 */

PG_FUNCTION_INFO_V1(uri_at_uri);
Datum uri_at_uri(PG_FUNCTION_ARGS)
{
    text	*uriText1,
		*uriText2;
    uri_t	*uriStruct;
    char	*uriString;

    if(PG_ARGISNULL(0))
	PG_RETURN_NULL();

    if(PG_ARGISNULL(1))
	PG_RETURN_TEXT_P(PG_GETARG_TEXT_P_COPY(0));
    
    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);

    uriText1 = PG_GETARG_TEXT_P(0);
    uriText2 = PG_GETARG_TEXT_P(1);

    uriStruct = uri_object(VARDATA(uriText2),VARATT_SIZEP(uriText2) - VARHDRSZ);
    uriStruct = uri_abs(uriStruct, VARDATA(uriText1), VARATT_SIZEP(uriText1) - VARHDRSZ);

    PG_FREE_IF_COPY(uriText1,0);
    PG_FREE_IF_COPY(uriText2,1);

    uriString = uri_uri(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_scheme);
Datum get_uri_scheme(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_scheme(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_user);
Datum get_uri_user(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_user(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_passwd);
Datum get_uri_passwd(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_passwd(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_host);
Datum get_uri_host(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_host(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_port);
Datum get_uri_port(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_port(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_path);
Datum get_uri_path(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_path(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_query);
Datum get_uri_query(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_query(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_frag);
Datum get_uri_frag(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_frag(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_robots);
Datum get_uri_robots(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uriString = uri_robots(uriStruct);
    if((uriString == NULL) || (strlen(uriString) == 0))
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_root);
Datum get_uri_root(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uri_user_set(uriStruct,NULL);
    uri_passwd_set(uriStruct,NULL);
    uri_path_set(uriStruct,NULL);
    uri_params_set(uriStruct,NULL);
    uri_query_set(uriStruct,NULL);
    uri_frag_set(uriStruct,NULL);

    if((uri_consistent(uriStruct) == 0) || ((uriString = uri_uri(uriStruct)) == NULL) || strlen(uriString) == 0)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_relative);
Datum get_uri_relative(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    uri_scheme_set(uriStruct,NULL);
    uri_user_set(uriStruct,NULL);
    uri_passwd_set(uriStruct,NULL);
    uri_host_set(uriStruct,NULL);
    uri_port_set(uriStruct,NULL);

    if((uri_consistent(uriStruct) == 0) || ((uriString = uri_uri(uriStruct)) == NULL) || strlen(uriString) == 0)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(get_uri_md5);
Datum get_uri_md5(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString;
    uri_t	*uriStruct;

    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_FURI_MD5 | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(VARDATA(uriText),VARATT_SIZEP(uriText) - VARHDRSZ);
    PG_FREE_IF_COPY(uriText,0);

    if((uri_consistent(uriStruct) == 0) || ((uriString = uri_furi(uriStruct)) == NULL) || strlen(uriString) == 0)
	PG_RETURN_NULL();

    PG_RETURN_TEXT_P(uriString_uriText(uriString));
}

PG_FUNCTION_INFO_V1(uri_isvalid);
Datum uri_isvalid(PG_FUNCTION_ARGS)
{
    text	*uriText = PG_GETARG_TEXT_P(0);
    char	*uriString = pg_do_encoding_conversion(VARDATA(uriText),VARSIZE(uriText) - VARHDRSZ,GetDatabaseEncoding(),PG_LATIN1);
    uri_t	*uriStruct;

    PG_FREE_IF_COPY(uriText,0);
    uri_mode_set(URI_MODE_URI_STRICT_SCHEME | URI_MODE_CANNONICAL | URI_MODE_LOWER_SCHEME);
    uriStruct = uri_object(uriString,strlen(uriString));
    pfree(uriString);

    if(uriStruct == NULL || uri_consistent(uriStruct) == 0)
	PG_RETURN_BOOL(false);

    PG_RETURN_BOOL(true);
}

/*
 *  Btree Operator Class Functions
 */

static
int uri_cmp_internal(text *arg1, text *arg2)
{
    int		result;
    int		len1,len2;

    len1 = VARSIZE(arg1) - VARHDRSZ;
    len2 = VARSIZE(arg2) - VARHDRSZ;

    result = strncmp(VARDATA(arg1), VARDATA(arg2), len1 < len2 ? len1 : len2);
    if ((result == 0) && (len1 != len2))
        result = (len1 < len2) ? -1 : 1;

    return result;
}

PG_FUNCTION_INFO_V1(bt_uri_cmp);
Datum bt_uri_cmp(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = uri_cmp_internal(arg1, arg2);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_INT32(result);
}

PG_FUNCTION_INFO_V1(uri_eq);
Datum uri_eq(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) == 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uri_ne);
Datum uri_ne(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) != 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uri_lt);
Datum uri_lt(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) < 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uri_le);
Datum uri_le(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) <= 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uri_gt);
Datum uri_gt(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) > 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}

PG_FUNCTION_INFO_V1(uri_ge);
Datum uri_ge(PG_FUNCTION_ARGS)
{
    text		*arg1,*arg2;
    int			result;

    arg1 = PG_GETARG_TEXT_P(0);
    arg2 = PG_GETARG_TEXT_P(1);

    result = (uri_cmp_internal(arg1, arg2) >= 0);
    PG_FREE_IF_COPY(arg1,0);
    PG_FREE_IF_COPY(arg2,1);

    PG_RETURN_BOOL(result);
}
